-----------------------------------------------------------------------------------------------
-- Client Lua Script for CooldownTimeline
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- CooldownTimeline Module Definition
-----------------------------------------------------------------------------------------------
local CooldownTimeline = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function CooldownTimeline:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.cooldownSpells = {} -- List of spells that are on cooldown
	self.maxSpellCooldownTime = 120 -- Max cooldown time that we want to see, this way we don't see stuff such as the path abilities
	self.combat = false
	
	-- These settings will get restored and saved
	self.cdtSettings = {}
	
	self.cdtSettings.lock = false -- Bar is locked
	self.cdtSettings.combat = false -- Bar is only visible in combat
	self.cdtSettings.opacity = 0.5

    return o
end

function CooldownTimeline:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- CooldownTimeline OnLoad
-----------------------------------------------------------------------------------------------
function CooldownTimeline:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("CooldownTimeline.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- CooldownTimeline OnDocLoaded
-----------------------------------------------------------------------------------------------
function CooldownTimeline:OnDocLoaded()

	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "CooldownTimelineForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
		-- if the xmlDoc is no longer needed, you should set it to nil
		-- self.xmlDoc = nil
		
		-- Register handlers for events, slash commands and timer, etc.
		-- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
		Apollo.RegisterSlashCommand("cdt", "OnOpenSettings", self)
		
		-- Do additional Addon initialization here
		Apollo.RegisterEventHandler("UnitEnteredCombat", "OnEnteredCombat", self)
		--Apollo.RegisterEventHandler("VarChange_FrameCount", "OnFrame", self)
		self.timer = ApolloTimer.Create(0.01, true, "OnFrame", self)
		
			self.test = false
		
		self:GetUnitVariables()
		self:ApplySettings()		
		
	    self.wndMain:Show(not self.cdtSettings.combat, true)
	end
end

function CooldownTimeline:ApplySettings()
	self.wndMain:SetStyle("Moveable", not self.cdtSettings.lock)
	self.wndMain:SetStyle("IgnoreMouse", self.cdtSettings.lock)
	
	if self.cdtSettings.combat and not self.combat then
		self.wndMain:Show(false, true)
	else
		self.wndMain:Show(true, true)
	end
	
	if self.cdtSettings.nOL then
		self.wndMain:SetAnchorOffsets(self.cdtSettings.nOL, self.cdtSettings.nOT, self.cdtSettings.nOR, self.cdtSettings.nOB)
	end
	

	self.wndMain:SetOpacity(self.cdtSettings.opacity)
	
	self.currentWidth = self.wndMain:GetWidth()
end

function CooldownTimeline:GetUnitVariables()
	self.playerUnit = GameLib.GetPlayerUnit()
		--Get class, fix for spellslinger API; TODO: remove when API is fixed
	if self.playerUnit ~= nil then
		local classId = self.playerUnit:GetClassId()
		self.className = GameLib.GetClassName(classId)
	end

end
-----------------------------------------------------------------------------------------------
-- CooldownTimeline Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here
function CooldownTimeline:OnEnteredCombat(unit)
	if not unit:IsThePlayer() then
		return
	end	
	
	self.combat = true
	
	if self.cdtSettings.combat then
		self.wndMain:Show(true, true)
	end
	
end

function CooldownTimeline:OnFrame()
	if self.playerUnit == nil then
		self:GetUnitVariables()
	end

	if self.cdtSettings.combat and not self.playerUnit:IsInCombat() then

		if self.wndMain:IsShown() then
			self.wndMain:Show(false, true)
		end
		
		if self.className ~= "Spellslinger" then
			return -- Shamefully, Spellslinger will currently have to suffer, since we will lose track of some cooldowns if we don't track them at all times. TODO: remove when API has been fixed
		end
	end

	--TODO: start this process when a spell has finished casting, check casting bar on how to do this

	--Get container
	local container = self.wndMain:FindChild("CooldownContainer")
	
	--Apparently dont have a choice but to put these here..
	self.abilities = AbilityBook.GetAbilitiesList()
	self.innate = GameLib.GetClassInnateAbilitySpells()
	
	if self.abilities == nil then
		return
	end

	local spellList = self:GetActiveSpellList()
	
	for idx = 1, #spellList do
		local spellObject = spellList[idx]
		
		--Add spell to cooldownlist
			if spellObject:GetCooldownRemaining() > 0 and self.cooldownSpells[spellObject:GetId()] == nil then
			--Add item
			local spellItem = Apollo.LoadForm(self.xmlDoc, "SpellItem", container, self)
			spellItem:FindChild("SpellIcon"):SetSprite(spellObject:GetIcon())
			spellItem:FindChild("SpellText"):SetText(math.floor(spellObject:GetCooldownRemaining()) .. "s")
							
			-- I've added the cooldown current as a fix for the Spellslinger class for now; TODO: remove this variable when the API is fixed
			self.cooldownSpells[spellObject:GetId()] = {object = spellObject, item = spellItem, cooldownCurrent = spellObject:GetCooldownRemaining(), cooldownStarted = os.clock()}		
		end
		
		if self.cooldownSpells[spellObject:GetId()] ~= nil then
			--Cooldown is on screen
			local spellItem = self.cooldownSpells[spellObject:GetId()]["item"]
			--Spellslinger API fixes; remove when API has been fixed
			local cooldownCurrent = self.cooldownSpells[spellObject:GetId()]["cooldownCurrent"]
			local cooldownStarted = self.cooldownSpells[spellObject:GetId()]["cooldownStarted"]
			
			local cooldownRemaining
			
			if self.className == "Spellslinger" and (spellObject:GetName() == "Charged Shot" or spellObject:GetName() == "Vitality Burst" or spellObject:GetName() == "Wild Barrage") then
				cooldownRemaining = cooldownCurrent - (os.clock() - cooldownStarted)
			else
				cooldownRemaining = spellObject:GetCooldownRemaining()
			end
			
			--Use for sorting
			--self.cooldownSpells[spellObject:GetId()]["cooldownRemaining"] = cooldownRemaining
						
			if cooldownRemaining <= 0 then
				local opacity = spellItem:GetOpacity()
				if spellItem:GetOpacity() == 1 then
					spellItem:SetOpacity(0)
				elseif spellItem:GetOpacity() == 0 then
					spellItem:Destroy()
					self.cooldownSpells[spellObject:GetId()] = nil
				end
			else
				local left,top,right,bottom = spellItem:GetAnchorOffsets()
				local cooldownWidth = container:GetWidth()
				
				local spellWidth = 50
				
				right  = -cooldownWidth - (-cooldownWidth * (cooldownRemaining / self:GetCurrentMaxCooldownTime(self.className))) + spellWidth		
				left = right - spellWidth
										
				spellItem:SetAnchorOffsets(left, top, right, bottom)
				spellItem:FindChild("SpellText"):SetText(math.floor(cooldownRemaining) .. "s")			

			end

		end
	end
end

--Added class argument as a Spellslinger API fix; TODO: remove when API has been fixed
function CooldownTimeline:GetCurrentMaxCooldownTime(class)
	local maxCooldownTime = 0
	
	for key,value in pairs(self.cooldownSpells) do
		local spellObject = value["object"]
		--API Fix objects; TODO: remove when API has been fixed
		local cooldownCurrent = value["cooldownCurrent"]
		local cooldownStarted = value["cooldownStarted"]
		
		if class == "Spellslinger" and (spellObject:GetName()== "Charged Shot" or spellObject:GetName()== "Vitality Burst" or spellObject:GetName()== "Wild Barrage") then
			local cooldownRemaining = cooldownCurrent - (os.clock() - cooldownStarted)
			
			if cooldownRemaining > 0 and cooldownRemaining > maxCooldownTime and spellObject:GetCooldownTime() <= self.maxSpellCooldownTime then
				maxCooldownTime = spellObject:GetCooldownTime()			
			end

		else
			if spellObject:GetCooldownRemaining() > 0 and spellObject:GetCooldownTime() > maxCooldownTime and spellObject:GetCooldownTime() <= self.maxSpellCooldownTime then
				maxCooldownTime = spellObject:GetCooldownTime()			
			end
		end
	end
	
	return maxCooldownTime
end

function CooldownTimeline:GetActiveSpellList()
	local spells = {}
	
	for idx = 1, #self.abilities do
		if self.abilities[idx].bIsActive then
			local currentTier = self.abilities[idx].tTiers[self.abilities[idx].nCurrentTier]
			local spellObject = currentTier.splObject
			
			table.insert(spells, spellObject)
		end
	end
	
	for idx = 1, #self.innate.tSpells  do
		table.insert(spells, self.innate.tSpells[idx])
	end
	
	return spells
end

function CooldownTimeline:OnOpenSettings()
	if self.settingsForm ~= nil then
		return
	end
	self.settingsForm = Apollo.LoadForm(self.xmlDoc, "CooldownTimelineSettings", nil, self)
	
	-- View the settings correctly
	local settingsPanel = self.settingsForm:FindChild("SettingsPanel")

	-- Lock	
	settingsPanel:FindChild("LockCheck"):SetCheck(self.cdtSettings.lock)
	settingsPanel:FindChild("CombatCheck"):SetCheck(self.cdtSettings.combat)
	settingsPanel:FindChild("WidthBar"):SetValue(self.wndMain:GetWidth())
	settingsPanel:FindChild("OpacityBar"):SetValue(self.cdtSettings.opacity*100)

	
	-- Show the window
	self.settingsForm:Show(true, true)
	
	-- Always show the bar regardless of anything
	self.timer:Stop()
	self.wndMain:Show(true, true)
end


---------------------------------------------------------------------------------------------------
-- CooldownTimelineSettings Functions
---------------------------------------------------------------------------------------------------

function CooldownTimeline:OnSaveClick()
	local settingsPanel = self.settingsForm:FindChild("SettingsPanel")
	
	-- Locked settings
	self.cdtSettings.lock = settingsPanel:FindChild("LockCheck"):IsChecked()
	self.cdtSettings.combat = settingsPanel:FindChild("CombatCheck"):IsChecked()
	
	local apL, apT, apR, apB = self.wndMain:GetAnchorOffsets()
	self.cdtSettings.nOL = apL
	self.cdtSettings.nOT = apT
	self.cdtSettings.nOR = apL + math.floor(settingsPanel:FindChild("WidthBar"):GetValue())
	self.cdtSettings.nOB = apB
	
	self.cdtSettings.opacity = settingsPanel:FindChild("OpacityBar"):GetValue()/100
	
	self:ApplySettings()
	-- Reshow since we still want to see the wndMain
	self.wndMain:Show(true, true)
end

function CooldownTimeline:OnCloseClick()
	-- Set the width of the timeline to the settings
	local apL, apT, apR, apB = self.wndMain:GetAnchorOffsets()
	self.wndMain:SetAnchorOffsets(apL, apT, apL + self.currentWidth, apB)
	
	self.wndMain:Show(not self.cdtSettings.combat, true)
	
	self.settingsForm:Close()
	self.timer:Start()
	
	self.settingsForm = nil
end


function CooldownTimeline:OnWidthChanged( wndHandler, wndControl, fNewValue, fOldValue )
	local apL, apT, apR, apB = self.wndMain:GetAnchorOffsets()
	self.wndMain:SetAnchorOffsets(apL, apT, apL + math.floor(fNewValue), apB)
end


function CooldownTimeline:OnOpacityChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.wndMain:SetOpacity(fNewValue/100)
end


--------------------------------
-- Save stuff
--------------------------------

function CooldownTimeline:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	local tSavedData = self.cdtSettings
	
	tSavedData.nOL, tSavedData.nOT, tSavedData.nOR, tSavedData.nOB = self.wndMain:GetAnchorOffsets()
	tSavedData.opacity = self.wndMain:GetOpacity()
	
	return tSavedData
end

function CooldownTimeline:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	if tSavedData ~= nil then
		--self.tUserSettings = tSavedData
		for idx, item in pairs(tSavedData) do
			self.cdtSettings[idx] = item
		end				
	end
end

-----------------------------------------------------------------------------------------------
-- CooldownTimeline Instance
-----------------------------------------------------------------------------------------------
local CooldownTimelineInst = CooldownTimeline:new()
CooldownTimelineInst:Init()
